@@ -29,8 +29,9 @@
 #include "slave.h"
 #include "misc.h"
 #include "xdmcp.h"
+#include "choose.h"
 
-static const gchar RCSid[]="$Id: display.c,v 1.11 2001/05/17 08:26:19 jirka Exp $";
+static const gchar RCSid[]="$Id: display.c,v 1.13 2001/06/22 22:24:27 jirka Exp $";
 
 /* External vars */
 extern gboolean GdmXdmcp;
@@ -154,7 +155,6 @@ gdm_display_manage (GdmDisplay *d)
     switch (d->slavepid = fork()) {
 
     case 0:
-
 	setpgid (0, 0);
 
 	/* Close XDMCP fd in slave process */
@@ -241,7 +241,7 @@ gdm_display_unmanage (GdmDisplay *d)
 void
 gdm_display_dispose (GdmDisplay *d)
 {
-    if (!d)
+    if (d == NULL)
 	return;
 
     if (d->type == TYPE_XDMCP) {
@@ -289,6 +289,10 @@ gdm_display_dispose (GdmDisplay *d)
     g_free (d->bcookie);
     d->bcookie = NULL;
 
+    if (d->indirect_id > 0)
+	    gdm_choose_indirect_dispose_empty_id (d->indirect_id);
+    d->indirect_id = 0;
+
     g_free (d);
 }
 
