@@ -35,8 +35,9 @@
 #include "xdmcp.h"
 #include "verify.h"
 #include "display.h"
+#include "errorgui.h"
 
-static const gchar RCSid[]="$Id: gdm.c,v 1.40 2001/06/02 08:57:09 jirka Exp $";
+static const gchar RCSid[]="$Id: gdm.c,v 1.49 2001/07/02 21:10:54 jirka Exp $";
 
 
 /* Local functions */
@@ -70,6 +71,7 @@ gchar *GdmLocaleFile = NULL;
 gchar *GdmGnomeDefaultSession = NULL;
 gchar *GdmAutomaticLogin = NULL;
 gboolean GdmAutomaticLoginEnable = FALSE;
+gboolean GdmAlwaysRestartServer = FALSE;
 gchar *GdmConfigurator = NULL;
 gboolean GdmConfigAvailable = FALSE;
 gboolean GdmSystemMenu = FALSE;
@@ -88,6 +90,7 @@ gchar *GdmXKeepsCrashing = NULL;
 gchar *GdmXKeepsCrashingConfigurators = NULL;
 gchar *GdmHalt = NULL;
 gchar *GdmReboot = NULL;
+gchar *GdmSuspend = NULL;
 gchar *GdmServAuthDir = NULL;
 gchar *GdmUserAuthDir = NULL;
 gchar *GdmUserAuthFile = NULL;
@@ -113,6 +116,7 @@ gboolean  GdmDebug = FALSE;
 gboolean  GdmVerboseAuth = FALSE;
 gboolean  GdmAllowRoot = FALSE;
 gboolean  GdmAllowRemoteRoot = FALSE;
+gboolean  GdmAllowRemoteAutoLogin = FALSE;
 gint  GdmRelaxPerms = 0;
 gint  GdmRetryDelay = 0;
 gchar *GdmTimedLogin = NULL;
@@ -120,8 +124,9 @@ gboolean GdmTimedLoginEnable = FALSE;
 gint GdmTimedLoginDelay = 0;
 
 /* set in the main function */
-static char **stored_argv = NULL;
-static int stored_argc = 0;
+char **stored_argv = NULL;
+int stored_argc = 0;
+char *stored_path = NULL;
 
 
 /**
@@ -153,6 +158,7 @@ gdm_config_parse (void)
     GdmDisplayInit = gnome_config_get_string (GDM_KEY_INITDIR);
     GdmAutomaticLoginEnable = gnome_config_get_bool (GDM_KEY_AUTOMATICLOGIN_ENABLE);
     GdmAutomaticLogin = gnome_config_get_string (GDM_KEY_AUTOMATICLOGIN);
+    GdmAlwaysRestartServer = gnome_config_get_bool (GDM_KEY_ALWAYSRESTARTSERVER);
     GdmGreeter = gnome_config_get_string (GDM_KEY_GREETER);
     GdmGroup = gnome_config_get_string (GDM_KEY_GROUP);
     GdmHalt = gnome_config_get_string (GDM_KEY_HALT);
@@ -177,6 +183,7 @@ gdm_config_parse (void)
     GdmDefaultLocale = gnome_config_get_string (GDM_KEY_LOCALE);
     GdmServAuthDir = gnome_config_get_string (GDM_KEY_SERVAUTH);
     GdmSessDir = gnome_config_get_string (GDM_KEY_SESSDIR);
+    GdmSuspend = gnome_config_get_string (GDM_KEY_SUSPEND);
     GdmLocaleFile = gnome_config_get_string (GDM_KEY_LOCFILE);
     GdmGnomeDefaultSession = gnome_config_get_string (GDM_KEY_GNOMEDEFAULTSESSION);
     GdmUser = gnome_config_get_string (GDM_KEY_USER);
@@ -190,6 +197,7 @@ gdm_config_parse (void)
 
     GdmAllowRoot = gnome_config_get_bool (GDM_KEY_ALLOWROOT);
     GdmAllowRemoteRoot = gnome_config_get_bool (GDM_KEY_ALLOWREMOTEROOT);
+    GdmAllowRemoteAutoLogin = gnome_config_get_bool (GDM_KEY_ALLOWREMOTEAUTOLOGIN);
     GdmRelaxPerms = gnome_config_get_int (GDM_KEY_RELAXPERM);
     GdmUserMaxFile = gnome_config_get_int (GDM_KEY_MAXFILE);
     GdmSessionMaxFile = gnome_config_get_int (GDM_KEY_SESSIONMAXFILE);
@@ -212,6 +220,13 @@ gdm_config_parse (void)
 
     /* sanitize some values */
 
+#ifndef HAVE_LIBXDMCP
+    if (GdmXdmcp) {
+	    gdm_info (_("gdm_config_parse: XDMCP was enabled while there is no XDMCP support, turning it off"));
+	    GdmXdmcp = FALSE;
+    }
+#endif
+
     if ( ! GdmAutomaticLoginEnable ||
 	gdm_string_empty (GdmAutomaticLogin)) {
 	    g_free (GdmAutomaticLogin);
@@ -238,9 +253,13 @@ gdm_config_parse (void)
 	    GdmTimedLogin = NULL;
     }
 
-    if (GdmTimedLoginDelay < 10) {
-	    gdm_info (_("gdm_config_parse: TimedLoginDelay less then 10, so I will just use 10."));
-	    GdmTimedLoginDelay = 10;
+    if (GdmTimedLoginDelay < 5) {
+	    gdm_info (_("gdm_config_parse: TimedLoginDelay less then 5, so I will just use 5."));
+	    GdmTimedLoginDelay = 5;
+    }
+
+    if (GdmMaxIndirect < 0) {
+	    GdmMaxIndirect = 0;
     }
 
     /* Prerequisites */ 
@@ -586,36 +605,37 @@ deal_with_x_crashes (GdmDisplay *d)
     if ( ! just_abort &&
 	access ("/usr/bin/open", X_OK) == 0) {
 	    char *dialog; /* do we have dialog?*/
+
+	    /* Shit if we knew what the program was to tell the user,
+	     * the above script would have been defined and we'd run
+	     * it for them */
+	    char *error = _("I cannot start the X server (your graphical "
+			    "interface).  It is likely that it is not set "
+			    "up correctly.  You will need to log in on a "
+			    "console and rerun the X configuration "
+			    "program.  Then restart GDM.");
+
 	    dialog = gnome_is_program_in_path ("dialog");
+	    if (dialog == NULL)
+		    dialog = gnome_is_program_in_path ("gdialog");
 	    if (dialog != NULL) {
 		    char *command = 
 			    g_strdup_printf
 			    ("/usr/bin/open -s -w -- /bin/sh -c 'clear ; "
 			     "%s --msgbox \"%s\" 10 70 ; clear'",
 			     dialog,
-			     _("I cannot start the X server (your graphical "
-			       "interface).  It is likely that it is not set "
-			       "up correctly.  You will need to log in on a "
-			       "console and rerun the X configuration "
-			       "program.  Then restart GDM."));
-		    /* Shit if we knew what the program was to tell the user,
-		     * the above script would have been defined and we'd run
-		     * it for them */
+			     error);
 		    system (command);
+		    g_free (command);
+		    g_free (dialog);
 	    } else {
 		    char *command = 
 			    g_strdup_printf
 			    ("/usr/bin/open -s -w -- /bin/sh -c 'clear ; "
-			     "echo \"%s\" 10 70' ; clear",
-			     _("I cannot start the X server (your graphical "
-			       "interface).  It is likely that it is not set "
-			       "up correctly.  You will need to log in on a "
-			       "console and rerun the X configuration "
-			       "program.  Then restart GDM."));
-		    /* Shit if we knew what the program was to tell the user,
-		     * the above script would have been defined and we'd run
-		     * it for them */
+			     "echo \"%s\" 10 70 ; read ; clear'",
+			     error);
 		    system (command);
+		    g_free (command);
 	    }
     } /* else {
        * At this point .... screw the user, we don't know how to
@@ -630,6 +650,26 @@ deal_with_x_crashes (GdmDisplay *d)
     return FALSE;
 }
 
+static gboolean
+bin_executable (const char *command)
+{
+	char **argv;
+
+	if (gdm_string_empty (command))
+		return FALSE;
+
+	argv = g_strsplit (command, argdelim, MAX_ARGS);	
+	if (argv != NULL &&
+	    argv[0] != NULL &&
+	    access (argv[0], X_OK) == 0) {
+		g_strfreev (argv);
+		return TRUE;
+	} else {
+		g_strfreev (argv);
+		return FALSE;
+	}
+}
+
 static void 
 gdm_cleanup_children (void)
 {
@@ -669,12 +709,39 @@ gdm_cleanup_children (void)
     }
 
     if (d->type != TYPE_LOCAL &&
-	(status == DISPLAY_REBOOT ||
+	(status == DISPLAY_RESTARTGDM ||
+	 status == DISPLAY_REBOOT ||
 	 status == DISPLAY_HALT)) {
-	    gdm_info (_("gdm_child_action: Reboot or Halt request from a non-local display %s"), d->name);
+	    gdm_info (_("gdm_child_action: Restart, Reboot or Halt request from a non-local display %s"), d->name);
 	    status = DISPLAY_REMANAGE;
     }
 
+    if (status == DISPLAY_CHOSEN) {
+	    /* forget about this indirect id, since this
+	     * display will be dead very soon, and we don't want it
+	     * to take the indirect display with it */
+	    d->indirect_id = 0;
+	    status = DISPLAY_REMANAGE;
+    }
+
+    /* checkout if we can actually do stuff */
+    switch (status) {
+    case DISPLAY_REBOOT:
+	    if ( ! bin_executable (GdmReboot))
+		    status = DISPLAY_REMANAGE;
+	    break;
+    case DISPLAY_HALT:
+	    if ( ! bin_executable (GdmHalt))
+		    status = DISPLAY_REMANAGE;
+	    break;
+    case DISPLAY_SUSPEND:
+	    if ( ! bin_executable (GdmSuspend))
+		    status = DISPLAY_REMANAGE;
+	    break;
+    default:
+	    break;
+    }
+
     /* Autopsy */
     switch (status) {
 	
@@ -706,6 +773,27 @@ gdm_cleanup_children (void)
 	gdm_error (_("gdm_child_action: Halt failed: %s"), strerror (errno));
 	break;
 
+    case DISPLAY_SUSPEND:	/* Suspend machine */
+	gdm_info (_("gdm_child_action: Master suspending..."));
+
+	final_cleanup ();
+
+	argv = g_strsplit (GdmReboot, argdelim, MAX_ARGS);	
+	execv (argv[0], argv);
+
+	gdm_error (_("gdm_child_action: Suspend failed: %s"), strerror (errno));
+	break;
+	
+
+    case DISPLAY_RESTARTGDM:
+	final_cleanup ();
+	if (stored_path != NULL)
+		putenv (stored_path);
+	execvp (stored_argv[0], stored_argv);
+	gdm_error (_("Failed to restart self"));
+	_exit (1);
+	break;
+
     case DISPLAY_XFAILED:       /* X sucks */
 	/* in remote case just drop to _REMANAGE */
 	if (d->type == TYPE_LOCAL) {
@@ -784,6 +872,8 @@ mainloop_sig_callback (gint8 sig, gpointer data)
 
     case SIGHUP:
       term_cleanup ();
+      if (stored_path != NULL)
+	      putenv (stored_path);
       execvp (stored_argv[0], stored_argv);
       gdm_error (_("Failed to restart self"));
       _exit (1);
@@ -813,6 +903,8 @@ store_argv (int argc, char *argv[])
 {
 	int i;
 
+	stored_path = g_strdup (g_getenv ("PATH"));
+
 	stored_argv = g_new0 (char *, argc + 1);
 	for (i = 0; i < argc; i++)
 		stored_argv[i] = g_strdup (argv[i]);
@@ -833,6 +925,20 @@ main (int argc, char *argv[])
     bindtextdomain(PACKAGE, GNOMELOCALEDIR);
     textdomain(PACKAGE);
 
+    /* This is an utter hack.  BUT we don't want to have the message in
+     * another program as it is supposed to be a last ditch effort to talk
+     * to the user, so what happens is that we respawn ourselves with
+     * an argument --run-error-dialog error dialog_type <x>:<y>:<width>:<height>
+     * the coordinates are the coordinates of this screen (when using
+     * xinerama) they can all be 0 */
+    if (argc == 5 &&
+	strcmp (argv[1], "--run-error-dialog") == 0) {
+	    int x = 0, y = 0, width = 0, height = 0;
+	    sscanf (argv[4], "%d:%d:%d:%d", &x, &y, &width, &height);
+	    gdm_run_errorgui (argv[2], argv[3], x, y, width, height);
+	    _exit (0);
+    }
+
     /* XDM compliant error message */
     if (getuid() != 0)
 	gdm_fail (_("Only root wants to run gdm\n"));
