@@ -56,6 +56,8 @@
 
 #include <config.h> 
 #include <gnome.h>
+
+#ifdef HAVE_LIBXDMCP
 #include <stdio.h>
 #include <stdlib.h>
 #include <syslog.h>
@@ -74,6 +76,8 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/utsname.h>
+#include <sys/ioctl.h>
+#include <net/if.h>
 #include <fcntl.h>
 
 #ifdef HAVE_TCPWRAPPERS
@@ -81,21 +85,27 @@
 #endif
 
 #include "gdm.h"
-#include "xdmcp.h"
 #include "display.h"
-#include "misc.h"
 #include "auth.h"
+#endif /* HAVE_LIBXDMCP */
 
-static const gchar RCSid[]="$Id: xdmcp.c,v 1.16 2001/06/02 08:57:09 jirka Exp $";
+#include "misc.h"
+#include "choose.h"
+#include "xdmcp.h"
+
+static const gchar RCSid[]="$Id: xdmcp.c,v 1.28 2001/07/03 09:13:33 jirka Exp $";
+gint pending = 0;
+
+#ifdef HAVE_LIBXDMCP
 
 /* TCP Wrapper syslog control */
 gint allow_severity = LOG_INFO;
 gint deny_severity = LOG_WARNING;
 
-gint pending = 0;
-
 static gint xdmcpfd = -1;
+static guint xdmcp_source = 0;
 static gint choosefd = -1;
+static guint choose_source = 0;
 static gint globsessid;
 static gchar *sysid;
 static ARRAY8 servhost;
@@ -114,28 +124,37 @@ extern gint GdmPort;		/* UDP port number */
 extern gboolean GdmIndirect;	/* Honor XDMCP_INDIRECT, i.e. choosing */
 extern gint GdmMaxIndirectWait;	/* Max wait between INDIRECT_QUERY and MANAGE */
 extern gint GdmDispPerHost;	/* Max number of displays per remote host */
+extern gchar *GdmTimedLogin;
+extern gboolean GdmAllowRemoteAutoLogin;
 
 extern gboolean GdmXdmcp;	/* xdmcp enabled */
 
 /* Local prototypes */
-static gboolean gdm_xdmcp_decode_packet (void);
-static void gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa, gint len, gint type);
-static void gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id, ARRAYofARRAY8Ptr authlist);
+static gboolean gdm_xdmcp_decode_packet (GIOChannel *source,
+					 GIOCondition cond,
+					 gpointer data);
+static void gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa,
+				    gint len,
+				    gint type);
+static void gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id,
+					  struct sockaddr_in *clnt_sa,
+					  struct in_addr *local_addr,
+					  ARRAYofARRAY8Ptr authlist);
 static void gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len);
 static void gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len);
 static void gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len);
 static void gdm_xdmcp_handle_keepalive (struct sockaddr_in *clnt_sa, gint len);
 static void gdm_xdmcp_send_willing (struct sockaddr_in *clnt_sa);
 static void gdm_xdmcp_send_unwilling (struct sockaddr_in *clnt_sa, gint type);
-static void gdm_xdmcp_send_accept (struct sockaddr_in *clnt_sa, gint displaynum);
+static void gdm_xdmcp_send_accept (const char *hostname, struct sockaddr_in *clnt_sa, gint displaynum);
 static void gdm_xdmcp_send_decline (struct sockaddr_in *clnt_sa);
 static void gdm_xdmcp_send_refuse (struct sockaddr_in *clnt_sa, CARD32 sessid);
 static void gdm_xdmcp_send_failed (struct sockaddr_in *clnt_sa, CARD32 sessid);
 static void gdm_xdmcp_send_alive (struct sockaddr_in *clnt_sa, CARD32 sessid);
 static gboolean gdm_xdmcp_host_allow (struct sockaddr_in *cnlt_sa);
-static GdmDisplay *gdm_xdmcp_display_alloc (struct sockaddr_in *, gint);
+static GdmDisplay *gdm_xdmcp_display_alloc (const char *hostname, gint);
 static GdmDisplay *gdm_xdmcp_display_lookup (CARD32 sessid);
-static void gdm_xdmcp_display_dispose_check (gchar *name);
+static void gdm_xdmcp_display_dispose_check (const gchar *name);
 static void gdm_xdmcp_displays_check (void);
 
 
@@ -161,23 +180,154 @@ static XdmAuthRec serv_authlist = {
     { (CARD16) 0, (CARD8 *) 0 }
 };
 
+static const GList *
+peek_local_address_list (void)
+{
+	static GList *the_list = NULL;
+	static time_t last_time = 0;
+	struct in_addr *addr;
+#ifdef SIOCGIFCONF
+	struct sockaddr_in *sin;
+	struct ifconf ifc;
+	struct ifreq *ifr;
+	char *buf;
+	int num;
+#else /* SIOCGIFCONF */
+	char hostbuf[BUFSIZ];
+	struct hostent *he;
+	int i;
+#endif
 
-extern gboolean gdm_choose_socket_handler (GIOChannel *source, GIOCondition cond, gint fd);
-extern GdmIndirectDisplay *gdm_choose_indirect_alloc (struct sockaddr_in *clnt_sa);
-extern GdmIndirectDisplay *gdm_choose_indirect_lookup (struct sockaddr_in *clnt_sa);
+	/* don't check more then every 5 seconds */
+	if (last_time + 5 > time (NULL))
+		return the_list;
+
+	g_list_foreach (the_list, (GFunc)g_free, NULL);
+	g_list_free (the_list);
+	the_list = NULL;
+
+	last_time = time (NULL);
+
+#ifdef SIOCGIFCONF
+#ifdef SIOCGIFNUM
+	if (ioctl (xdmcpfd, SIOCGIFNUM, &num) < 0) {
+		num = 64;
+	}
+#else
+	num = 64;
+#endif
+
+	ifc.ifc_len = sizeof(struct ifreq) * num;
+	ifc.ifc_buf = buf = g_malloc (ifc.ifc_len);
+	if (ioctl (xdmcpfd, SIOCGIFCONF, &ifc) < 0) {
+		gdm_error (_("%s: Cannot get local addresses!"),
+			   "peek_local_address_list");
+		return NULL;
+	}
+
+	ifr = ifc.ifc_req;
+	num = ifc.ifc_len / sizeof(struct ifreq);
+	for (; num-- > 0; ifr++) {
+		if (ioctl (xdmcpfd, SIOCGIFFLAGS, ifr) < 0 ||
+		    ! (ifr->ifr_flags & IFF_UP))
+			continue;
+#ifdef IFF_UNNUMBERED
+		if (ifr->ifr_flags & IFF_UNNUMBERED)
+			continue;
+#endif
+		if (ioctl (xdmcpfd, SIOCGIFADDR, ifr) < 0)
+			continue;
+
+		sin = (struct sockaddr_in *)&ifr->ifr_addr;
+
+		if (sin->sin_family != AF_INET ||
+		    sin->sin_addr.s_addr == INADDR_ANY ||
+		    sin->sin_addr.s_addr == INADDR_LOOPBACK)
+			continue;
+		addr = g_new0 (struct in_addr, 1);
+		memcpy (addr, &(sin->sin_addr.s_addr),
+			sizeof (struct in_addr));
+		the_list = g_list_append (the_list, addr);
+	}
+
+	g_free (buf);
+#else /* SIOCGIFCONF */
+	/* host based fallback, will likely only get 127.0.0.1 i think */
+
+	if (gethostname (hostbuf, BUFSIZ-1) != 0) {
+		gdm_error (_("%s: Could not get server hostname: %s!"),
+			   "peek_local_address_list",
+			   strerror (errno));
+		return NULL;
+	}
+	he = gethostbyname (hostbuf);
+	if (he == NULL) {
+		gdm_error (_("%s: Could not get address from hostname!"),
+			   "peek_local_address_list");
+		return NULL;
+	}
+	for (i = 0; he->h_addr_list[i] != NULL; i++) {
+		struct in_addr *laddr = (struct in_addr *)he->h_addr_list[i];
+
+		addr = g_new0 (struct in_addr, 1);
+		memcpy (addr, laddr, sizeof (struct in_addr));
+		the_list = g_list_append (the_list, addr);
+	}
+#endif
+
+	return the_list;
+}
 
 
-int
+static gboolean
+is_loopback_addr (struct in_addr *ia)
+{
+	const char lo[] = {127,0,0,1};
+
+	if (ia->s_addr == INADDR_LOOPBACK ||
+	    memcmp (&ia->s_addr, lo, 4) == 0) {
+		return TRUE;
+	} else {
+		return FALSE;
+	}
+}
+
+static gboolean
+is_local_addr (struct in_addr *ia)
+{
+	const char lo[] = {127,0,0,1};
+
+	if (ia->s_addr == INADDR_LOOPBACK ||
+	    memcmp (&ia->s_addr, lo, 4) == 0) {
+		return TRUE;
+	} else {
+		const GList *list = peek_local_address_list ();
+
+		while (list != NULL) {
+			struct in_addr *addr = list->data;
+
+			if (memcmp (&ia->s_addr, &addr->s_addr, 4) == 0) {
+				return TRUE;
+			}
+
+			list = list->next;
+		}
+
+		return FALSE;
+	}
+}
+
+gboolean
 gdm_xdmcp_init (void)
 {
-    struct sockaddr_in serv_sa;
+    struct sockaddr_in serv_sa = {0};
     gchar hostbuf[256];
     struct utsname name;
     
     globsessid = time (NULL);
     
     /* Fetch and store local hostname in XDMCP friendly format */
-    if (gethostname (hostbuf, 255)) {
+    if (gethostname (hostbuf, 255) != 0) {
 	gdm_error (_("gdm_xdmcp_init: Could not get server hostname: %s!"), strerror (errno));
 	GdmXdmcp = FALSE;
 	return FALSE;
@@ -217,19 +367,13 @@ gdm_xdmcp_init (void)
 
 	fifopath = g_strconcat (GdmServAuthDir, "/.gdmchooser", NULL);
 
-	if (!fifopath) {
-	    gdm_error (_("gdm_xdmcp_init: Can't alloc fifopath"));
-	    gdm_xdmcp_close ();
-	    GdmXdmcp = FALSE;
-	    return FALSE;
-	}
-
 	unlink (fifopath);
 
 	if (mkfifo (fifopath, 0660) < 0) {
 	    gdm_error (_("gdm_xdmcp_init: Could not make FIFO for chooser"));
 	    gdm_xdmcp_close ();
 	    GdmXdmcp = FALSE;
+	    g_free (fifopath);
 	    return FALSE;
 	}
 
@@ -239,6 +383,7 @@ gdm_xdmcp_init (void)
 	    gdm_error (_("gdm_xdmcp_init: Could not open FIFO for chooser"));
 	    gdm_xdmcp_close ();
 	    GdmXdmcp = FALSE;
+	    g_free (fifopath);
 	    return FALSE;
 	}
 
@@ -254,38 +399,47 @@ gdm_xdmcp_init (void)
 void
 gdm_xdmcp_run (void)
 {
-    GIOChannel *xdmcpchan;
-    
-    xdmcpchan = g_io_channel_unix_new (xdmcpfd);
-    g_io_add_watch_full (xdmcpchan, G_PRIORITY_DEFAULT,
+	GIOChannel *xdmcpchan;
+
+	xdmcpchan = g_io_channel_unix_new (xdmcpfd);
+	xdmcp_source = g_io_add_watch_full
+		(xdmcpchan, G_PRIORITY_DEFAULT,
+		 G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
+		 gdm_xdmcp_decode_packet, NULL, NULL);
+	g_io_channel_unref (xdmcpchan);
+
+	if (GdmIndirect) {
+		GIOChannel *choosechan;
+
+		choosechan = g_io_channel_unix_new (choosefd);
+		choose_source = g_io_add_watch_full
+			(choosechan, G_PRIORITY_DEFAULT,
 			 G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
-			 (GIOFunc) gdm_xdmcp_decode_packet,
-			 GINT_TO_POINTER (xdmcpfd), NULL);
-    g_io_channel_unref (xdmcpchan);
-
-    if (GdmIndirect) {
-	GIOChannel *choosechan;
-
-	choosechan = g_io_channel_unix_new (choosefd);
-	g_io_add_watch_full (choosechan, G_PRIORITY_DEFAULT,
-			     G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
-			     (GIOFunc) gdm_choose_socket_handler,
-			     GINT_TO_POINTER (choosefd), NULL);
-	g_io_channel_unref (choosechan);
-    }
+			 gdm_choose_socket_handler, NULL, NULL);
+		g_io_channel_unref (choosechan);
+	}
 }
 
 
 void
 gdm_xdmcp_close (void)
 {
+	if (xdmcp_source > 0) {
+		g_source_remove (xdmcp_source);
+		xdmcp_source = 0;
+	}
+
 	if (xdmcpfd > 0) {
 		close (xdmcpfd);
 		xdmcpfd = -1;
 	}
 
-	if (GdmIndirect &&
-	    choosefd > 0) {
+	if (choose_source > 0) {
+		g_source_remove (choose_source);
+		choose_source = 0;
+	}
+
+	if (choosefd > 0) {
 		close (choosefd);
 		choosefd = -1;
 	}
@@ -293,7 +447,7 @@ gdm_xdmcp_close (void)
 
 
 static gboolean
-gdm_xdmcp_decode_packet (void)
+gdm_xdmcp_decode_packet (GIOChannel *source, GIOCondition cond, gpointer data)
 {
     struct sockaddr_in clnt_sa;
     gint sa_len = sizeof (clnt_sa);
@@ -320,9 +474,9 @@ gdm_xdmcp_decode_packet (void)
 	return TRUE;
     }
 
-    gdm_debug ("gdm_xdmcp_decode: Received opcode %s from client %s", 
-	       opcode_names[header.opcode], inet_ntoa (clnt_sa.sin_addr));
-
+    if (header.opcode <= ALIVE)
+	    gdm_debug ("gdm_xdmcp_decode: Received opcode %s from client %s", 
+		       opcode_names[header.opcode], inet_ntoa (clnt_sa.sin_addr));
 
     switch (header.opcode) {
 	
@@ -364,13 +518,12 @@ gdm_xdmcp_decode_packet (void)
     return TRUE;
 }
 
-
 static void 
 gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa, gint len, gint type)
 {
     ARRAYofARRAY8 clnt_authlist;
     gint i = 0, explen = 1;
-    
+
     gdm_debug ("gdm_xdmcp_handle_query: Opcode %d from %s", 
 	       type, inet_ntoa (clnt_sa->sin_addr));
     
@@ -388,9 +541,10 @@ gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa, gint len, gint type)
     
     if (len != explen) {
 	gdm_error (_("gdm_xdmcp_handle_query: Error in checksum")); 
+	XdmcpDisposeARRAYofARRAY8 (&clnt_authlist);
 	return;
     }
-    
+
     /* Check with tcp_wrappers if client is allowed to access */
     if (gdm_xdmcp_host_allow (clnt_sa)) {
 
@@ -398,47 +552,119 @@ gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa, gint len, gint type)
  	 * the pending list. If found send a FORWARD_QUERY to the
  	 * chosen manager. Otherwise alloc a new indirect display. */
 
-	if (GdmIndirect && type==INDIRECT_QUERY) {
-	    GdmIndirectDisplay *id = gdm_choose_indirect_lookup (clnt_sa);
+	if (GdmIndirect &&
+	    type == INDIRECT_QUERY) {
+		GdmIndirectDisplay *id = gdm_choose_indirect_lookup (clnt_sa);
 
-	    if (id) 
-		if (id->acctime + GdmMaxIndirectWait < time (NULL)) /* Expired? */
-		    gdm_xdmcp_send_forward_query (id, &clnt_authlist);
-		else
-		    gdm_xdmcp_send_unwilling (clnt_sa, type);
-	    else
-		gdm_choose_indirect_alloc (clnt_sa);
+		if (id != NULL &&
+		    id->chosen_host != NULL) {
+			/* if user chose us, then just send willing */
+			if (is_local_addr (id->chosen_host)) {
+				/* get rid of indirect, so that we don't get
+				 * the chooser */
+				gdm_choose_indirect_dispose (id);
+				gdm_xdmcp_send_willing (clnt_sa);
+			} else if (is_loopback_addr (&(clnt_sa->sin_addr))) {
+				/* woohoo! fun, I have no clue how to get
+				 * the correct ip, SO I just send forward
+				 * queries with all the different IPs */
+				const GList *list = peek_local_address_list ();
+
+				while (list != NULL) {
+					struct in_addr *addr = list->data;
+					
+					if ( ! is_loopback_addr (addr)) {
+						/* forward query to
+						 * chosen host */
+						gdm_xdmcp_send_forward_query
+							(id, clnt_sa, addr,
+							 &clnt_authlist);
+					}
+
+					list = list->next;
+				}
+			} else {
+				/* or send forward query to chosen host */
+				gdm_xdmcp_send_forward_query
+					(id, clnt_sa,
+					 &(clnt_sa->sin_addr),
+					 &clnt_authlist);
+			}
+		} else if (id == NULL) {
+			id = gdm_choose_indirect_alloc (clnt_sa);
+			if (id != NULL) {
+				gdm_xdmcp_send_willing (clnt_sa);
+			}
+		} else  {
+			gdm_xdmcp_send_willing (clnt_sa);
+		}
+	} else {
+		gdm_xdmcp_send_willing (clnt_sa);
 	}
-
-	gdm_xdmcp_send_willing (clnt_sa);
+    } else if (type == QUERY) {
+	    /* unwilling is ONLY sent for direct queries, never for broadcast
+	     * nor indirects */
+	    gdm_xdmcp_send_unwilling (clnt_sa, type);
     }
-    else
-	gdm_xdmcp_send_unwilling (clnt_sa, type);
 
     /* Dispose authlist from remote display */
-    /* XdmcpDisposeARRAYofARRAY8 (&clnt_authlist); */
+    XdmcpDisposeARRAYofARRAY8 (&clnt_authlist);
 }
 
-
 static void
-gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id, ARRAYofARRAY8Ptr authlist)
+gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id,
+			      struct sockaddr_in *clnt_sa,
+			      struct in_addr *local_addr,
+			      ARRAYofARRAY8Ptr authlist)
 {
-    ARRAY8 status;
-    XdmcpHeader header;
-    
-    status.data = sysid;
-    status.length = strlen (sysid);
-    
-    header.opcode = (CARD16) WILLING;
-    header.length = 6 + serv_authlist.authentication.length;
-    header.length += servhost.length + status.length;
-    header.version = XDM_PROTOCOL_VERSION;
-    XdmcpWriteHeader (&buf, &header);
-    
-    XdmcpWriteARRAY8 (&buf, &serv_authlist.authentication); /* Hardcoded authentication */
-    XdmcpWriteARRAY8 (&buf, &servhost);
-    XdmcpWriteARRAY8 (&buf, &status);
-    /* XdmcpFlush (xdmcpfd, &buf, clnt_sa, sizeof (struct sockaddr_in)); */
+	struct sockaddr_in sock = {0};
+	XdmcpHeader header;
+	int i, authlen;
+	ARRAY8 address;
+	ARRAY8 port;
+
+	g_assert (id != NULL);
+	g_assert (id->chosen_host != NULL);
+
+	gdm_debug ("gdm_xdmcp_send_forward_query: Sending forward query to %s",
+		   inet_ntoa (*id->chosen_host));
+	gdm_debug ("gdm_xdmcp_send_forward_query: Query contains %s:%d", 
+		   inet_ntoa (*local_addr),
+		   (int) ntohs (clnt_sa->sin_port));
+
+	authlen = 1;
+	for (i = 0 ; i < authlist->length ; i++) {
+		authlen += 2 + authlist->data[i].length;
+	}
+
+	/* we depend on this being 2 elsewhere as well */
+	port.length = 2;
+	port.data = g_new (char, 2);
+	memcpy (port.data, &(clnt_sa->sin_port), 2);
+
+	address.length = sizeof (struct in_addr);
+	address.data = (void *)g_new (struct in_addr, 1);
+	memcpy (address.data, local_addr, sizeof (struct in_addr));
+
+	header.opcode = (CARD16) FORWARD_QUERY;
+	header.length = authlen;
+	header.length += 2 + address.length;
+	header.length += 2 + port.length;
+	header.version = XDM_PROTOCOL_VERSION;
+	XdmcpWriteHeader (&buf, &header);
+
+	XdmcpWriteARRAY8 (&buf, &address);
+	XdmcpWriteARRAY8 (&buf, &port);
+	XdmcpWriteARRAYofARRAY8 (&buf, authlist);
+
+	sock.sin_family = AF_INET;
+	sock.sin_port = htons (XDM_UDP_PORT);
+	sock.sin_addr.s_addr = id->chosen_host->s_addr;
+	XdmcpFlush (xdmcpfd, &buf, (XdmcpNetaddr) &sock,
+		    (int)sizeof (struct sockaddr_in));
+
+	g_free (port.data);
+	g_free (address.data);
 }
 
 
@@ -449,32 +675,25 @@ gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len)
     ARRAY8 clnt_port;
     ARRAYofARRAY8 clnt_authlist;
     gint i = 0, explen = 1;
-    static struct sockaddr_in *disp_sa;
-    guint port = 0;
-    struct in_addr ia;
-    
-    gdm_debug ("gdm_xdmcp_handle_forward_query: ForwardQuery from %s", 
-	       inet_ntoa (clnt_sa->sin_addr));
+    struct sockaddr_in disp_sa = {0};
     
     /* Read display address */
     if (! XdmcpReadARRAY8 (&buf, &clnt_addr)) {
-	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
 	gdm_error (_("gdm_xdmcp_handle_forward_query: Could not read display address"));
 	return;
     }
     
     /* Read display port */
     if (! XdmcpReadARRAY8 (&buf, &clnt_port)) {
-	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
-	XdmcpDisposeARRAYofARRAY8 (&clnt_port);
+	XdmcpDisposeARRAY8 (&clnt_addr);
 	gdm_error (_("gdm_xdmcp_handle_forward_query: Could not read display port number"));
 	return;
     }
     
     /* Extract array of authentication names from Xdmcp packet */
     if (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_authlist)) {
-	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
-	XdmcpDisposeARRAYofARRAY8 (&clnt_port);
+	XdmcpDisposeARRAY8 (&clnt_addr);
+	XdmcpDisposeARRAY8 (&clnt_port);
 	gdm_error (_("gdm_xdmcp_handle_forward_query: Could not extract authlist from packet")); 
 	return;
     }
@@ -494,34 +713,32 @@ gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len)
 	goto out;
     }
     
+    if (clnt_port.length != 2 ||
+	clnt_addr.length != 4) {
+	    gdm_error (_("gdm_xdmcp_handle_forward_query: Bad address")); 
+	    goto out;
+    }
+
+    disp_sa.sin_family = AF_INET;
+
     /* Find client port number */
-    for (i = 0 ; i < clnt_port.length ; i++)
-	port = port*256+clnt_port.data[i];
+    memcpy (&disp_sa.sin_port, clnt_port.data, 2);
     
-    /* Find client address. Ugly, ugly. Endianness sucks... */
-    memmove (&ia.s_addr, clnt_addr.data, MIN(clnt_addr.length, sizeof(ia.s_addr)));
+    /* Find client address */
+    memcpy (&disp_sa.sin_addr.s_addr, clnt_addr.data, 4);
     
-    gdm_debug ("gdm_xdmcp_handle_forward_query: Got FORWARD_QUERY from display: %s, port %d", 
-	       inet_ntoa (ia), port);
-    
-    /* Assemble sockaddr_in struct to pass on */
-    disp_sa = g_new0 (struct sockaddr_in, 1);
-    disp_sa->sin_family = AF_INET;
-    disp_sa->sin_port = htons (port);
-    disp_sa->sin_addr.s_addr = ia.s_addr;
+    gdm_debug ("gdm_xdmcp_handle_forward_query: Got FORWARD_QUERY for display: %s, port %d", 
+	       inet_ntoa (disp_sa.sin_addr), ntohs (disp_sa.sin_port));
     
     /* Check with tcp_wrappers if display is allowed to access */
-    if (gdm_xdmcp_host_allow (disp_sa)) 
-	gdm_xdmcp_send_willing (disp_sa);
-    else
-	gdm_xdmcp_send_unwilling (disp_sa, FORWARD_QUERY);
+    if (gdm_xdmcp_host_allow (&disp_sa)) 
+	    gdm_xdmcp_send_willing (&disp_sa);
 
   out:
-    g_free(disp_sa);
     /* Cleanup */
     XdmcpDisposeARRAYofARRAY8 (&clnt_authlist);
-    XdmcpDisposeARRAYofARRAY8 (&clnt_port);
-    XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
+    XdmcpDisposeARRAY8 (&clnt_port);
+    XdmcpDisposeARRAY8 (&clnt_addr);
 }
 
 
@@ -545,7 +762,7 @@ gdm_xdmcp_send_willing (struct sockaddr_in *clnt_sa)
     XdmcpWriteARRAY8 (&buf, &serv_authlist.authentication); /* Hardcoded authentication */
     XdmcpWriteARRAY8 (&buf, &servhost);
     XdmcpWriteARRAY8 (&buf, &status);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, sizeof (struct sockaddr_in));
+    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
 }
 
 static void
@@ -568,9 +785,27 @@ gdm_xdmcp_send_unwilling (struct sockaddr_in *clnt_sa, gint type)
     
     XdmcpWriteARRAY8 (&buf, &servhost);
     XdmcpWriteARRAY8 (&buf, &status);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, sizeof (struct sockaddr_in));
+    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
 }
 
+static char *
+get_host_from_addr (struct sockaddr_in *clnt_sa)
+{
+	char *hostname;
+	struct hostent *he;
+
+	/* Find client hostname */
+	he = gethostbyaddr ((gchar *) &clnt_sa->sin_addr,
+			    sizeof (struct in_addr),
+			    AF_INET);
+
+	if (he != NULL) {
+		hostname = g_strdup (he->h_name);
+	} else {
+		hostname = g_strdup (inet_ntoa (clnt_sa->sin_addr));
+	}
+	return hostname;
+}
 
 static void
 gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
@@ -611,24 +846,34 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
     /* This is TCP/IP - we don't care */
     if (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_addr)) {
 	gdm_error (_("gdm_xdmcp_handle_request: Could not read Client Address"));
+        XdmcpDisposeARRAY16 (&clnt_conntyp);
 	return;
     }
     
     /* Read authentication type */
     if (! XdmcpReadARRAY8 (&buf, &clnt_authname)) {
 	gdm_error (_("gdm_xdmcp_handle_request: Could not read Authentication Names"));
+	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
+	XdmcpDisposeARRAY16 (&clnt_conntyp);
 	return;
     }
     
     /* Read authentication data */
     if (! XdmcpReadARRAY8 (&buf, &clnt_authdata)) {
 	gdm_error (_("gdm_xdmcp_handle_request: Could not read Authentication Data"));
+	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
+	XdmcpDisposeARRAY16 (&clnt_conntyp);
+	XdmcpDisposeARRAY8 (&clnt_authname);
 	return;
     }
     
     /* Read and select from supported authorization list */
     if (! XdmcpReadARRAYofARRAY8 (&buf, &clnt_authorization)) {
 	gdm_error (_("gdm_xdmcp_handle_request: Could not read Authorization List"));
+	XdmcpDisposeARRAY8 (&clnt_authdata);
+	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
+	XdmcpDisposeARRAY16 (&clnt_conntyp);
+	XdmcpDisposeARRAY8 (&clnt_authname);
 	return;
     }
     
@@ -640,6 +885,11 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
     /* Manufacturer ID */
     if (! XdmcpReadARRAY8 (&buf, &clnt_manufacturer)) {
 	gdm_error (_("gdm_xdmcp_handle_request: Could not read Manufacturer ID"));
+	XdmcpDisposeARRAY8 (&clnt_authname);
+	XdmcpDisposeARRAY8 (&clnt_authdata);
+	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
+	XdmcpDisposeARRAYofARRAY8 (&clnt_authorization);
+	XdmcpDisposeARRAY16 (&clnt_conntyp);
 	return;
     }
     
@@ -659,19 +909,40 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
     if (explen != len) {
 	gdm_error (_("gdm_xdmcp_handle_request: Failed checksum from %s"),
 		   inet_ntoa (clnt_sa->sin_addr));
+	XdmcpDisposeARRAY8 (&clnt_authname);
+	XdmcpDisposeARRAY8 (&clnt_authdata);
+	XdmcpDisposeARRAY8 (&clnt_manufacturer);
+	XdmcpDisposeARRAYofARRAY8 (&clnt_addr);
+	XdmcpDisposeARRAYofARRAY8 (&clnt_authorization);
+	XdmcpDisposeARRAY16 (&clnt_conntyp);
 	return;
     }
     
     gdm_debug ("gdm_xdmcp_handle_request: pending=%d, MaxPending=%d, sessions=%d, MaxSessions=%d",
 	       pending, GdmMaxPending, sessions, GdmMaxSessions);
-    
+
     /* Check if ok to manage display */
     if (mitauth &&
-	pending < GdmMaxPending && 
-	sessions < GdmMaxSessions)
-	gdm_xdmcp_send_accept (clnt_sa, clnt_dspnum);
-    else
-	gdm_xdmcp_send_decline (clnt_sa);	
+	sessions < GdmMaxSessions) {
+	    char *disp;
+	    char *hostname = get_host_from_addr (clnt_sa);
+	    disp = g_strdup_printf ("%s:%d", hostname, clnt_dspnum);
+
+	    /* Check if we are already talking to this host */
+	    gdm_xdmcp_display_dispose_check (disp);
+	    g_free (disp);
+
+	    if (pending >= GdmMaxPending) {
+		    gdm_debug ("gdm_xdmcp_handle_request: maximum pending");
+		    gdm_xdmcp_send_decline (clnt_sa);	
+	    } else {
+		    gdm_xdmcp_send_accept (hostname, clnt_sa, clnt_dspnum);
+	    }
+
+	    g_free (hostname);
+    } else {
+	    gdm_xdmcp_send_decline (clnt_sa);	
+    }
 
     XdmcpDisposeARRAY8 (&clnt_authname);
     XdmcpDisposeARRAY8 (&clnt_authdata);
@@ -683,7 +954,9 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
 
 
 static void
-gdm_xdmcp_send_accept (struct sockaddr_in *clnt_sa, gint displaynum)
+gdm_xdmcp_send_accept (const char *hostname,
+		       struct sockaddr_in *clnt_sa,
+		       gint displaynum)
 {
     XdmcpHeader header;
     ARRAY8 authentype;
@@ -692,7 +965,7 @@ gdm_xdmcp_send_accept (struct sockaddr_in *clnt_sa, gint displaynum)
     ARRAY8 authdata;
     GdmDisplay *d;
     
-    d = gdm_xdmcp_display_alloc (clnt_sa, displaynum);
+    d = gdm_xdmcp_display_alloc (hostname, displaynum);
     
     authentype.data = (CARD8 *) 0;
     authentype.length = (CARD16) 0;
@@ -722,7 +995,7 @@ gdm_xdmcp_send_accept (struct sockaddr_in *clnt_sa, gint displaynum)
     XdmcpWriteARRAY8 (&buf, &authname);
     XdmcpWriteARRAY8 (&buf, &authdata);
     
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, sizeof (struct sockaddr_in));
+    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
     
     gdm_debug ("gdm_xdmcp_send_accept: Sending ACCEPT to %s with SessionID=%d", 
 	       inet_ntoa (clnt_sa->sin_addr), d->sessionid);
@@ -760,7 +1033,7 @@ gdm_xdmcp_send_decline (struct sockaddr_in *clnt_sa)
     XdmcpWriteARRAY8 (&buf, &authentype);
     XdmcpWriteARRAY8 (&buf, &authendata);
     
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, sizeof (struct sockaddr_in));
+    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
 }
 
 
@@ -771,6 +1044,7 @@ gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len)
     CARD16 clnt_dspnum;
     ARRAY8 clnt_dspclass;
     GdmDisplay *d;
+    GdmIndirectDisplay *id;
     gint logfd;
     
     gdm_debug ("gdm_xdmcp_handle_manage: Got MANAGE from %s", inet_ntoa (clnt_sa->sin_addr));
@@ -799,13 +1073,33 @@ gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len)
     
     /* Display Class */
     if (! XdmcpReadARRAY8 (&buf, &clnt_dspclass)) {
-        XdmcpDisposeARRAY8(&clnt_dspclass);
 	gdm_error (_("gdm_xdmcp_handle_manage: Could not read Display Class"));
 	return;
     }
     
     d = gdm_xdmcp_display_lookup (clnt_sessid);
-    
+    if (GdmIndirect) {
+	    id = gdm_choose_indirect_lookup (clnt_sa);
+
+	    /* This was an indirect thingie and nothing was yet chosen,
+	     * use a chooser */
+	    if (d != NULL &&
+		d->dispstat == XDMCP_PENDING &&
+		id != NULL &&
+		id->chosen_host == NULL) {
+		    d->use_chooser = TRUE;
+		    d->indirect_id = id->id;
+	    } else {
+		    d->indirect_id = 0;
+		    d->use_chooser = FALSE;
+		    if (id != NULL)
+			    gdm_choose_indirect_dispose (id);
+	    }
+    } else {
+	    d->indirect_id = 0;
+	    d->use_chooser = FALSE;
+    }
+
     if (d && d->dispstat == XDMCP_PENDING) {
 	gchar *logfile;
 
@@ -827,7 +1121,7 @@ gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len)
 	d->dispstat = XDMCP_MANAGED;
 	sessions++;
 	pending--;
-	
+
 	/* Start greeter/session */
 	if (!gdm_display_manage (d)) {
 	    gdm_xdmcp_send_failed (clnt_sa, clnt_sessid);
@@ -860,7 +1154,7 @@ gdm_xdmcp_send_refuse (struct sockaddr_in *clnt_sa, CARD32 sessid)
     
     XdmcpWriteHeader (&buf, &header);  
     XdmcpWriteCARD32 (&buf, sessid);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, sizeof (struct sockaddr_in));
+    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
 }
 
 
@@ -882,7 +1176,7 @@ gdm_xdmcp_send_failed (struct sockaddr_in *clnt_sa, CARD32 sessid)
     XdmcpWriteHeader (&buf, &header);
     XdmcpWriteCARD32 (&buf, sessid);
     XdmcpWriteARRAY8 (&buf, &status);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, sizeof (struct sockaddr_in));
+    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
 }
 
 
@@ -932,7 +1226,7 @@ gdm_xdmcp_send_alive (struct sockaddr_in *clnt_sa, CARD32 sessid)
     XdmcpWriteHeader (&buf, &header);
     XdmcpWriteCARD8 (&buf, 1);
     XdmcpWriteCARD32 (&buf, sessid);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, sizeof (struct sockaddr_in));
+    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
 }
 
 
@@ -940,6 +1234,12 @@ static gboolean
 gdm_xdmcp_host_allow (struct sockaddr_in *clnt_sa)
 {
 #ifdef HAVE_TCPWRAPPERS
+	
+    /* avoids a warning, my tcpd.h file doesn't include this prototype, even
+     * though the library does include the function and the manpage mentions it
+     */
+    extern int hosts_ctl (char *daemon, char *client_name, char *client_addr, char *client_user);
+
     struct hostent *client_he;
     gchar *client;
     
@@ -948,7 +1248,7 @@ gdm_xdmcp_host_allow (struct sockaddr_in *clnt_sa)
 			       sizeof (struct in_addr),
 			       AF_INET);
     
-    client = (client_he && client_he->h_name) ? client_he->h_name : NULL;
+    client = (client_he != NULL) ? client_he->h_name : NULL;
     
     /* Check with tcp_wrappers if client is allowed to access */
     return (hosts_ctl ("gdm", client ? client : "unknown", inet_ntoa (clnt_sa->sin_addr), ""));
@@ -959,10 +1259,9 @@ gdm_xdmcp_host_allow (struct sockaddr_in *clnt_sa)
 
 
 static GdmDisplay *
-gdm_xdmcp_display_alloc (struct sockaddr_in *clnt_sa, gint displaynum)
+gdm_xdmcp_display_alloc (const char *hostname, gint displaynum)
 {
     GdmDisplay *d = NULL;
-    struct hostent *client_he = NULL;
     
     d = g_new0 (GdmDisplay, 1);
     d->authfile = NULL;
@@ -982,25 +1281,16 @@ gdm_xdmcp_display_alloc (struct sockaddr_in *clnt_sa, gint displaynum)
     d->dispnum = displaynum;
 
     d->sleep_before_run = 0;
-    
-    /* Find client hostname */
-    client_he = gethostbyaddr ((gchar *) &clnt_sa->sin_addr,
-			       sizeof (struct in_addr),
-			       AF_INET);
-    
-    if (client_he) {
-	d->name = g_strdup_printf ("%s:%d", client_he->h_name, 
-				   displaynum);
-	d->hostname = g_strdup (client_he->h_name);
-    }
-    else {
-	d->name = g_strdup_printf ("%s:%d", inet_ntoa (clnt_sa->sin_addr),
-				   displaynum);
-	d->hostname = g_strdup (inet_ntoa (clnt_sa->sin_addr));
+    if (GdmAllowRemoteAutoLogin &&
+	! gdm_string_empty (GdmTimedLogin)) {
+	    d->timed_login_ok = TRUE;
+    } else {
+	    d->timed_login_ok = FALSE;
     }
     
-    /* Check if we are already talking to this host */
-    gdm_xdmcp_display_dispose_check (d->name);
+    d->name = g_strdup_printf ("%s:%d", hostname,
+			       displaynum);
+    d->hostname = g_strdup (hostname);
     
     /* Secure display with cookie */
     if (! gdm_auth_secure_display (d))
@@ -1040,7 +1330,7 @@ gdm_xdmcp_display_lookup (CARD32 sessid)
 
 
 static void
-gdm_xdmcp_display_dispose_check (gchar *name)
+gdm_xdmcp_display_dispose_check (const gchar *name)
 {
 	GSList *dlist;
 
@@ -1054,8 +1344,12 @@ gdm_xdmcp_display_dispose_check (gchar *name)
 		GdmDisplay *d = dlist->data;
 
 		if (d != NULL &&
+		    d->type == TYPE_XDMCP &&
 		    strcmp (d->name, name) == 0) {
-			gdm_display_dispose (d);
+			if (d->dispstat == XDMCP_MANAGED)
+				gdm_display_unmanage (d);
+			else
+				gdm_display_dispose (d);
 
 			/* restart as the list is now fucked */
 			dlist = displays;
@@ -1093,5 +1387,29 @@ gdm_xdmcp_displays_check (void)
 	}
 }
 
+#else /* HAVE_LIBXDMCP */
+
+/* Here come some empty stubs for no XDMCP support */
+int
+gdm_xdmcp_init  (void)
+{
+	gdm_error (_("gdm_xdmcp_init: No XDMCP support"));
+	return FALSE;
+}
+
+void
+gdm_xdmcp_run (void)
+{
+	gdm_error (_("gdm_xdmcp_run: No XDMCP support"));
+}
+
+void
+gdm_xdmcp_close (void)
+{
+	gdm_error (_("gdm_xdmcp_close: No XDMCP support"));
+}
+
+#endif /* HAVE_LIBXDMCP */
+
 
 /* EOF */
